#coding : utf - 8

#In[32]:

from lbmpy.session import *
from scipy.ndimage.filters import gaussian_filter

from pygrandchem_tests.config_anisotropic import get_system
from pystencils.datahandling import SerialDataHandling
from pygrandchem.grandchem_generation import *
from pystencils import show_code
import pycuda.driver as drv
from pystencils.simp.liveness_opts import *
from pystencils.simp.liveness_opts_exp import *
from pygrandchem.initialization import *
from pygrandchem.chemicalpotential import free_energy_from_config_object, FreeEnergy
from pystencils.boundaries import *

from pystencils import show_code
import pycuda.driver as drv
from pystencils.simp.liveness_opts import *
from pystencils.simp.liveness_opts_exp import *

domain_size = (512, 512, 128)
periodicity = (True, True, False)
fast_simplex_projection = True
optimization = { 'gpu_indexing_params': {"block_size": (32, 4, 2)}}
config = get_system(dim=len(domain_size))

phases = config['Parameters']['phases']
components = config['Parameters']['components']
diffusion_matrices = config['Parameters']['diffusion']
free_energy = config['FreeEnergy']

#Adding fields
dh = create_data_handling(domain_size, periodicity=periodicity, default_target=optimization['target'])
f = dh.fields
phi_src = dh.add_array('phi_src', values_per_cell=phases, layout='fzyx', latex_name='phi_s')
mu_src = dh.add_array('mu_src', values_per_cell=components, layout='fzyx', latex_name="mu_s")
mu_stag = dh.add_array('mu_stag', values_per_cell=(dh.dim, components), layout='f')
phi_dst = dh.add_array_like('phi_dst', 'phi_src')
mu_dst = dh.add_array_like('mu_dst', 'mu_src')

c = dh.add_array('c', values_per_cell=components, layout='fzyx', gpu=False)

mu_vanilla_eqs = create_mu_update_equations(phi_src, phi_dst, mu_src, mu_dst, free_energy, diffusion_matrices,
                                            config['Parameters'])

phi_vanilla_eqs = create_phi_update_equations(
    phi_src, phi_dst, mu_src, free_energy, config['Parameters'], simplex_projection=fast_simplex_projection)

init_boxes(dh, height=0.2)
initialize_concentration_field(dh, free_energy, config['Parameters']['initial_concentration'])
smooth_fields(dh, sigma=0.4, iterations=5, dim=dh.dim)
dh.synchronization_function(['phi_src', 'phi_dst', 'mu_src', 'mu_dst'])()
print(dh)


def bench_kernels(mu_kernel, phi_kernel):

    start = drv.Event()
    end = drv.Event()

    dh.run_kernel(mu_kernel)
    start.record()
    dh.run_kernel(mu_kernel)
    dh.run_kernel(mu_kernel)
    end.record()
    end.synchronize()
    msec = start.time_till(end) / 2
    print("mu_kernel: {:5.3f} ms".format(msec))

    dh.run_kernel(phi_kernel)
    start.record()
    dh.run_kernel(phi_kernel)
    dh.run_kernel(phi_kernel)
    end.record()
    end.synchronize()
    msec = start.time_till(end) / 2
    print("phi_kernel: {:5.3f} ms".format(msec))


sched_options = []

sched_options.append(option_none)
sched_options.append(option_none)
sched_options.append(option_none)
#sched_options.append(option_reschedule)
sched_options.append(option_liveness_opt_transformation)
#sched_options.append(option_dupl_reschedule)
sched_options.append(option_liveness_opt_transformation_shmem)
sched_options.append(option_liveness_opt_transformation_shmem2)
#sched_options.append(option_reschedule_shmem)
#sched_options.append(option_liveness_opt_transformation_shmem)
#sched_options.append(optionFuseSubs)
#sched_options.append(optionFuseFMAs)
#sched_options.append(optionFuseBoth)
#sched_options.append(optionRescheduleAtomize)
#sched_options.append(optionRescheduleAtomizeScramble)
#sched_options.append(optionDuplAtomizeReschedule)
#sched_options.append(optionDuplAtomizeRefuseReschedule)
#sched_options.append(optionDuplRescheduleAtomize)
#sched_options.append(optionDuplRescheduleAtomizeScramble)
#sched_options.append(optionSchedIteration)
#sched_options.append(optionAtomizeRescheduleNoSqrt)
#sched_options.append(optionAtomizeRescheduleNoDiv)
#sched_options.append(optionAtomizeRescheduleNoSqrtDiv)
#sched_options.append(optionAtomizeRescheduleNoPiecewise)
#sched_options.append(optionAtomizeRescheduleNoAll)

mu_rescheduled_eqs = sched_options[0](mu_vanilla_eqs)
phi_rescheduled_eqs = sched_options[0](phi_vanilla_eqs)

mu_kernel = create_kernel(mu_rescheduled_eqs, target="gpu", gpu_indexing_params={"block_size": (32, 4, 2)}).compile()

phi_kernel = create_kernel(phi_rescheduled_eqs, target="gpu", gpu_indexing_params={"block_size": (32, 4, 2)}).compile()

print("mu_kernel: " + str(mu_kernel.num_regs) + " regs")
print("phi_kernel: " + str(phi_kernel.num_regs) + " regs")

bench_kernels(mu_kernel, phi_kernel)

print()

dh.swap('mu_src', 'mu_dst')
dh.swap('phi_src', 'phi_dst')

for sched_option in sched_options:
    print(sched_option.__name__)
    mu_rescheduled_eqs = sched_option(mu_vanilla_eqs)
    phi_rescheduled_eqs = sched_option(phi_vanilla_eqs)

    mu_kernel = create_kernel(
        mu_rescheduled_eqs, target="gpu", gpu_indexing_params={
            "block_size": (32, 4, 2)
        }).compile()

    phi_kernel = create_kernel(
        phi_rescheduled_eqs, target="gpu", gpu_indexing_params={
            "block_size": (32, 4, 2)
        }).compile()

    print("mu_kernel: " + str(mu_kernel.num_regs) + " regs")
    print("phi_kernel: " + str(phi_kernel.num_regs) + " regs")

    bench_kernels(mu_kernel, phi_kernel)

    print()
